/*
    Meganizer - Media Organzier is a media management software
    Copyright (C) 2010  EngSaS - Engineering Solutions and Services Langenbach. All rights reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "dbsettings.h"

#include "sqllogin.h"
#include "msqldatabase.h"

#include <QSettings>
#include <QSqlQuery>
#include <QSqlError>

DBSettings::DBSettings ( SetupWizardObject *newObject, QWidget *parent, Qt::WindowFlags fl )
    : SetupWizardWidget ( newObject, parent, fl )
{
  setupUi ( this );
  setName ( tr ( "Database Settings" ) );

	if(!newObject)
		createDatabaseBox->setVisible(false);
	
  // load settings
  QStringList drivers = QSqlDatabase::drivers();
  // strip out don't supported drivers
  QStringList delSQLite = drivers.filter ( "sqlite", Qt::CaseInsensitive );
  foreach ( QString string, delSQLite ) {
    drivers.removeAll ( string );
  }

  typeEdit->addItems ( drivers );
  QSettings settings;
  settings.beginGroup ( "database" );
  typeEdit->setCurrentIndex ( typeEdit->findText ( settings.value ( "type" ).toString() ) );
  hostEdit->setText ( settings.value ( "host" ).toString() );
  dbEdit->setText ( settings.value ( "database" ).toString() );
	sslEdit->setChecked( settings.value( "ssl" ).toBool() );
  settings.endGroup();

  connect ( typeEdit, SIGNAL ( currentIndexChanged ( QString ) ), this, SLOT ( typeChanged ( QString ) ) );
  typeChanged ( typeEdit->currentText() );
}

bool DBSettings::finish()
{
  if ( !saveSettings() )
    return false;

  if ( !SqlLogin::checkSettings() ) {
    emit error ( tr ( "The database settings are wrong!" ) );
    return false;
  }

	// if create user or create database is selected,
	// we must use an admin connection
	QString userName = createUserNameEdit->text();
	QString password = createUserPasswordEdit->text();
	if(createDBEdit->isChecked() || createUserBox->isChecked()){
		QString dbName = dbEdit->text();

		QString adminConnName = "admin";
		SqlLogin adminLogin(tr("Administration login"), adminConnName, this);
		connectByPass(&adminLogin);
		if( adminLogin.exec() != QDialog::Accepted){
			emit error(tr("Could not connect to database server with administrator privileges."));
			return false;
		}
		MSqlDatabase db = MSqlDatabase::database(adminConnName);
		if(!db.open()){
			emit error(tr("Could not connect to database server with administrator privileges."));
			return false;
		}
		QSqlQuery query(db);
		// first create user if needed
		if(createUserBox->isChecked()){
			QString host = "%";
			if(userName.isEmpty()){
				emit error(tr("Empty Username is not allowed."));
				return false;
			}
			if(password.isEmpty()){
				emit error(tr("Empty password is not allowed."));
				return false;
			}

			// check wether user already exists
			// this query can not be used again (I think because of queryNotReuseable.first() later on)
			QSqlQuery queryNotReuseable(db);
			queryNotReuseable.prepare("SELECT User FROM mysql.user WHERE Host=:Host AND User=:User");
			queryNotReuseable.bindValue(":Host", host);
			queryNotReuseable.bindValue(":User", userName);
			if(!queryNotReuseable.exec()){
				emit error(tr("Could not check wether user exist."));
				db.close();
				return false;
			}
			if(queryNotReuseable.first()){
				// update password only, user exists already
				query.prepare("SET PASSWORD FOR ?@? = PASSWORD(?)");
				query.addBindValue(userName);
				query.addBindValue(host);
				query.addBindValue(password);
				if(!query.exec()){
					emit error(tr("Could not change password."));
					return false;
				}
			}
			else{
				// create new user
				query.prepare("CREATE USER ?@? IDENTIFIED BY ?");
				query.addBindValue(userName);
				query.addBindValue(host);
				query.addBindValue(password);
				if(!query.exec()){
					emit error(tr("Could not create user."));
					return false;
				}
			}
			// grant privileges
			query.prepare(QString("GRANT ALL ON %1.* TO ?@?").arg(dbName));
			query.addBindValue(userName);
			query.addBindValue(host);
			if(!query.exec()){
				emit error(tr("Could not grant privileges to choosen database."));
				return false;
			}
			if(!query.exec("FLUSH PRIVILEGES")){
				emit error(tr("Could not reload privileges."));
				return false;
			}
		}

		// create database
		if(createDBEdit->isChecked()){
			// check wether database already exists
			if(!query.exec(QString("CREATE DATABASE IF NOT EXISTS %1").arg(dbName))){
				emit error(tr("Could not create database."));
				return false;
			}
		}
		db.close();
	}

  SqlLogin *sqlLogin = new SqlLogin(this);
	sqlLogin->loadAndCheckSettings(userName, "", "", "", password);
 	connectByPass ( sqlLogin );

	bool result = false;
  if ( sqlLogin->exec() == QDialog::Accepted )
		result =  true;
	delete sqlLogin;
	
  return result;
}

bool DBSettings::saveSettings ()
{
	if(dbEdit->text().contains(QRegExp("[;,\\s]+"))){
		emit error(tr("Database name contains not allowed strings"));
		return false;
	}
	
  QSettings settings;
  settings.beginGroup ( "database" );
  settings.setValue ( "type", typeEdit->currentText() );
  settings.setValue ( "host", hostEdit->text() );
  settings.setValue ( "database", dbEdit->text() );
	settings.setValue ( "ssl", sslEdit->isChecked() );
  settings.endGroup();
	emit reloadSettings();
  return true;
}

void DBSettings::typeChanged ( QString text )
{
  if ( text.contains ( "SQLITE" ) ) {
    dbLabel->setText ( tr ( "File:" ) );
    hostEdit->setVisible ( false );
    hostLabel->setVisible ( false );
  }

  else {
    dbLabel->setText ( tr ( "Database:" ) );
    hostEdit->setVisible ( true );
    hostLabel->setVisible ( true );
  }
}

#include "dbsettings.moc"
