/*
    Meganizer - Media Organzier is a media management software
    Copyright (C) 2010  EngSaS - Engineering Solutions and Services Langenbach. All rights reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "mediamodel.h"

#include "searchplugin.h"
#include "mwidget.h"
#include "delegateforeignkey.h"
#include "delegateuser.h"
#include "searchdetailswidget.h"
#include "progressstatuswidget.h"
#include "querythread.h"
#include "msqldatabase.h"

#include <QSqlQuery>
#include <QSqlError>
#include <QAbstractItemView>
#include <QTableView>
#include <QSqlRelationalDelegate>
#include <QSettings>
#include <QMessageBox>
#include <QTextTable>
#include <QDateTime>

MediaModel::MediaModel ( QObject* parent ) : QAbstractItemModel ( parent )
{
	ownThread = false;
	compatibilityMode = false;
}

MediaModel::MediaModel ( QueryThread* queryThread, QObject* parent ) : QAbstractItemModel ( parent )
{
	ownThread = false;
	compatibilityMode = true;
   setupModel(queryThread);
}

MediaModel::~MediaModel()
{
  qDeleteAll ( media );
  media.clear();
	if(ownThread){
		// stopping thread
		// cancel query first
		thread->cancelQuery();
		// stop eventloop
		thread->quit();
		// wait 2000 seconds to stop the eventloop
		thread->wait(2000);
		delete thread;
	}
}

QString MediaModel::group()
{
  return myGroup;
}

QString MediaModel::name()
{
  return myName;
}

MediaModel* MediaModel::createNewModel()
{
  return new MediaModel(0, 0);
}

Medium *MediaModel::newMedium() const
{
  return new Medium();
}

Medium* MediaModel::addEmptyMedium()
{
  Medium *medium = newMedium();
	doAddMedium(medium);
  return medium;
}

bool MediaModel::addMedium ( unsigned int materialsId )
{
  Medium *medium = addEmptyMedium();

  if ( !medium )
    return false;

  if ( medium->select ( materialsId ) )
    return true;

	emit layoutAboutToBeChanged();
  media.removeAll ( medium );
	emit layoutChanged();

  return false;
}

bool MediaModel::addMedium ( Medium *medium )
{
  if ( !medium )
    return false;

  doAddMedium(medium);

  return true;
}

bool MediaModel::addMedium ( QSqlRecord record )
{
  Medium *medium = addEmptyMedium();
  return medium->select ( record );
}

Medium* MediaModel::mediumFromIndex ( QModelIndex index ) const
{
//   if ( !index.isValid() ) {
//     return 0;
//   }
//   else {
//     return static_cast<Medium *> ( index.internalPointer() );
//   }

	// otherwise it seems that there can be valid indexes without valid pointer
	// Steps to reproduce: Search -> select some media -> export
	if(index.row() < 0)
		return 0;
	if(index.row() >= media.size())
		return 0;
	
	return media[index.row()];
}

QModelIndex MediaModel::indexFromMedium ( Medium *medium )
{
  int row = media.indexOf ( medium );

  if ( row >= 0 )
    return index ( row, 0 );

  // check childs here
  foreach ( Medium *tmp, media ) {
    row = tmp->childs().indexOf ( medium );

    if ( row >= 0 )
      return index ( row, 0, indexFromMedium ( tmp ) );
  }

  return QModelIndex();
}


bool MediaModel::removeMedium ( Medium* medium )
{
  int index = media.indexOf ( medium );

  if ( index < 0 )
    return false;

  media.removeOne ( medium );

  return true;
}

bool MediaModel::removeMedium ( QModelIndex index )
{
  if ( !index.isValid() )
    return false;

  Medium *medium = mediumFromIndex ( index );

  if ( !medium )
    return false;

  return removeMedium ( medium );
}

bool MediaModel::store()
{
  bool result = true;
  foreach ( Medium *medium, media ) {
    if ( !medium->store() )
      result = false;
  }

  return false;
}

bool MediaModel::load ( QList< SqlConstraint > constraints, QList< SearchPlugin* > plugins,
												int userId, SearchDetailsWidget* searchDetailsWidget)
{
	// if the thread is needed somewhere else than in this function
	// (searchOtherUsers and searchSearchPlugins use the thread created here),
	// than delete this thread part and comment out the same part at setupModel()
	if(!thread){
		thread = new QueryThread();
		ownThread = true;
		connect(thread, SIGNAL(started()), this, SLOT(setupThread()));
	}
	if(!thread->isStarted())
		thread->start();
	connect(thread, SIGNAL(result(QSqlRecord)), this, SLOT(addMedium(QSqlRecord)));
	
	lastLoad.lastSearchPlugins = plugins;
	lastLoad.rawConstraints = constraints;
  lastUserId = userId;
	lastLoad.table = "viewmaterials";
	lastLoad.order = "ORDER BY title";
	lastLoad.usedDetailsWidget = false;
	lastLoad.needPersons = false;
	lastLoad.wish = false;
	lastLoad.projects = false;
	lastLoad.gifts = false;
	lastLoad.userNotNeeded = false;

	// search only for handled media types
  if ( mediaCondition.isValid() )
    constraints.append ( mediaCondition );
	foreach(SqlConstraint constraint, constraints){
		if(constraint.needPerson())
			lastLoad.needPersons = true;
		if(constraint.constraint().contains( "wish=1" ))
			lastLoad.wish = true;
		if(constraint.constraint() == "literatur" ){
			lastLoad.projects = true;
			constraints.removeAll(constraint);
		}
		if(constraint.constraint().contains( "presenteeId" ))
			lastLoad.gifts = true;
	}
	
	// check wether external search is possible
  bool searchExternal = false;
  // if there is a constraint, we can search for a title or EAN or ...
  // in that case we can do a external search
  if ( (constraints.size() > 0) && (mediaCondition.isValid()) ) {
    // an external search should be only done, no valid userId is present
    // and the default order is choosen (means userId = 0)
    // All Local is userId == -1 and searchExternal must be false
		// userId == -2 means all local and external
    if ( (userId == 0) || (userId == -2) )
      searchExternal = true;
  }
  if ( lastLoad.wish || lastLoad.projects || lastLoad.gifts )
    searchExternal = false;
	lastLoad.searchExternal = searchExternal;
	
		// check which table should be searched
	if(searchDetailsWidget){
		QList<SqlConstraint> typeConstraints =  searchDetailsWidget->constraints();
		if(typeConstraints.size() > 0){
			lastLoad.table = searchDetailsWidget->searchTable(lastLoad.needPersons);
			constraints << typeConstraints;
			lastLoad.usedDetailsWidget = true;
		}		
	}
	
	bool personsSupported = true;
	if(lastLoad.projects){
		personsSupported = false;
		lastLoad.table = "viewprojectmaterials";
		lastLoad.order = "GROUP BY title";
	}
	else if(lastLoad.gifts){
		personsSupported = false;
		lastLoad.table = "viewgiftmaterials";
		// prevent the insertion of another users constraint
		lastLoad.userNotNeeded = true;
	}
	
	if((lastLoad.needPersons) && (personsSupported)){
		if(!lastLoad.usedDetailsWidget){
			lastLoad.table = "viewmaterialpersons";
		}
		else{
			personsSupported = searchDetailsWidget->personsSupported();
		}
		
	}
	
	if(!personsSupported){
		// we cannot search for persons, so remove all constraints concerning persons
		bool removedConstraints = false;
		foreach(SqlConstraint constraint, constraints){
			if(constraint.needPerson()){
				constraints.removeAll(constraint);
				removedConstraints = true;
			}
		}
		if(removedConstraints){
			emit warning(tr("Your search combination can not be combined with person search,\n"
											" so the person conditions are removed from query!"));
		}
	}
	
	foreach(SqlConstraint constraint, constraints){
		if(!constraint.isValid())
			continue;
		lastLoad.strings << constraint.constraint();
		lastLoad.bindValues << constraint.bindValues();
	}

	ProgressStatusWidget *statusWidget = new ProgressStatusWidget();
	statusWidget->setLabel(tr("Local: "));
	connect(statusWidget, SIGNAL(removeStatusWidget(QWidget*)), this, SLOT(removeStatusWidgetByPass(QWidget*)));
	connect(thread, SIGNAL(setRange(int,int)), statusWidget, SLOT(setRange(int,int)));
	connect(thread, SIGNAL(setValue(int)), statusWidget, SLOT(setValue(int)));
	connect(thread, SIGNAL(finished(bool)), statusWidget, SLOT(removeMe()));
	emit showStatusWidget(statusWidget);
	// search for local media of other users and afterwards search searchplugins
	connect(thread, SIGNAL(finished(bool)), this, SLOT(searchOtherUsers(bool)));
	
	if(!thread->waitForStart()){
		thread->disconnect(this);
		disconnect(thread);
		statusWidget->removeMe();
		return false;
	}
	bool result = true;
  // if userId > 0, a user is choosen, so search only the (local) media
  // which belongs to him
  if ( userId > 0 ) {
		if(!lastLoad.userNotNeeded){
			lastLoad.strings << QString ( "usersId=?" );
			lastLoad.bindValues << userId;
		}
    QString where = " WHERE " + lastLoad.strings.join ( " AND " );
		thread->execQuery(QString ( "SELECT * FROM %1%2 " ).arg(lastLoad.table).arg ( where ), lastLoad.bindValues);
  }

  // All local media is choosen
  else if ( userId == -1 ) {
    QString where;

    if ( lastLoad.strings.size() > 0 )
      where = " WHERE " + lastLoad.strings.join ( " AND " );

		thread->execQuery(QString ( "SELECT * FROM %1%2 %3" ).arg(lastLoad.table).arg ( where ).arg(lastLoad.order), lastLoad.bindValues);
  }

  // the default order is choosen
  else {
    // first search local media which belongs the actual user
		MSqlDatabase db = MSqlDatabase::database();
		bool closeDB = !db.isOpen();
		if(closeDB)
			db.open();
		QSqlQuery query;
    lastLoad.actualUserId = 0;
		query.prepare ( "SELECT id FROM users WHERE userName=:userName" );
		query.bindValue(":userName", db.userName());
    query.exec ( );

    if ( !query.first() ) {
      emit warning ( tr ( "Could not find id of actual user!\n%1\nError: %2\nQuery: %3" )
                     .arg ( db.userName() )
                     .arg ( query.lastError().text() )
                     .arg ( query.lastQuery() )
                   );
    }
    else {
      lastLoad.actualUserId = query.record().value ( "id" ).toUInt();
      QStringList userConstraints = lastLoad.strings;
			QList<QVariant> userConstraintBindValues = lastLoad.bindValues;
			if(!lastLoad.userNotNeeded){
				userConstraints << QString ( "usersId=?" );
				userConstraintBindValues << lastLoad.actualUserId;
			}
      QString where = " WHERE " + userConstraints.join ( " AND " );
			thread->execQuery(QString ( "SELECT * FROM %1%2 %3" ).arg(lastLoad.table).arg ( where ).arg(lastLoad.order), userConstraintBindValues);
    }
    if(closeDB)
			db.close();
	}
	return result;
}

bool MediaModel::reload()
{
// 	if(!clear())
// 		return false;
// 	return load(lastConstraints, lastSearchPlugins, lastUserId);
  return true;
}

void MediaModel::configureView ( QAbstractItemView* view )
{
	if(!view->model())
		view->setModel(this);
	// set selection mode
	view->setSelectionBehavior(QAbstractItemView::SelectRows);
	view->setSelectionMode(QAbstractItemView::ExtendedSelection);
	
	// add delegates
	addItemDelegates(view);
	
	// hide and resize columns
	QTableView *tableView = qobject_cast<QTableView*>(view);
	if(tableView)
		hideAndResizeColumns(tableView);
}

void MediaModel::addItemDelegates ( QAbstractItemView* view )
{
  if ( !view )
    return;

  Medium *medium = newMedium();

  if ( !medium )
    return;

  QStringList attrs = medium->supportedAttributes();

  view->setItemDelegateForColumn ( medium->attrs("type").column, new DelegateForeignKey ( "materialtypes", "id", "name", view ) );

  view->setItemDelegateForColumn ( medium->attrs("listPriceCurrency").column,
																	 new DelegateForeignKey ( "currencies", "id", "shortName", view ) );

  view->setItemDelegateForColumn ( medium->attrs("owner").column,  new DelegateUser ( view ) );

  delete medium;
}

void MediaModel::hideAndResizeColumns ( QTableView *view )
{
  if ( !view )
    return;
	
  // hide columns
	for ( int i = 0; i < hideColumns.size(); i++ ) {
    view->hideColumn( hideColumns[i]);
  }

  // resize columns
  // resize all visible columns without biggest column
  int usedWidth = 0;

  for ( int i = 0; i < columnCount(); i++ ) {
    if ( ( !hideColumns.contains ( i ) ) && ( i != biggestColumn ) ) {
      view->resizeColumnToContents ( i );
      usedWidth += view->columnWidth ( i );
    }
  }

  // get width, which can be used for the title column
	// leave some pixels free at the right side
	int minWidth = 100;
	usedWidth += 50;
  if ( (usedWidth + minWidth) < view->width() ) {
    // some space left to grow title column
    view->setColumnWidth ( biggestColumn, view->width() - usedWidth );
  }
  else{
		view->setColumnWidth ( biggestColumn, minWidth );
	}
  view->sortByColumn(biggestColumn);
}

QVariant MediaModel::data ( const QModelIndex& index, int role ) const
{
  if ( ( role != Qt::DisplayRole ) && ( role != Qt::EditRole ) && ( role != Qt::ForegroundRole ) && ( role != Qt::DecorationRole ) )
    return QVariant();

  Medium *medium = mediumFromIndex ( index );

  if ( !medium )
    return QVariant();

  int col = index.column();

  return medium->value ( col, role );
}

int MediaModel::columnCount ( const QModelIndex& parent ) const
{
  Medium *medium = mediumFromIndex ( parent );

  if ( medium )
    return  medium->columnCount();

  if ( media.size() > 0 )
    return media[0]->columnCount();

	Medium *med = newMedium();
	int count = med->columnCount();
	delete med;
  return count;
}

int MediaModel::rowCount ( const QModelIndex& parent ) const
{
  Medium *medium = mediumFromIndex ( parent );

  if ( !medium )
    return media.size();

  return medium->rowCount();
}

QModelIndex MediaModel::parent ( const QModelIndex& child ) const
{
  Medium *medium = mediumFromIndex ( child );

  if ( !medium )
    return QModelIndex();

  Medium *mParent = medium->parent();

  if ( !mParent )
    return QModelIndex();

  int row = media.indexOf ( mParent );

  if ( row > -1 )
    return createIndex ( row, 0, mParent );

  Medium *mGrandParent = mParent->parent();

  if ( !mGrandParent )
    return QModelIndex();

  row = mGrandParent->childs().indexOf ( mParent );

  if ( row > -1 )
    return createIndex ( row, 0, mParent );

  return QModelIndex();
}

QModelIndex MediaModel::index ( int row, int column, const QModelIndex& parent ) const
{
  if ( ( row < 0 ) || ( column < 0 ) )
    return QModelIndex();

  Medium *parentMedium = mediumFromIndex ( parent );

  if ( !parentMedium ) {
    if ( row < media.size() )
      return createIndex ( row, column, media[row] );

    return QModelIndex();
  }

  QList<Medium*> childs = parentMedium->childs();

  if ( row < childs.size() )
    return createIndex ( row, column, childs[row] );

  return QModelIndex();
}

bool MediaModel::hasChildren ( const QModelIndex& parent ) const
{
  Medium *parentMedium = mediumFromIndex ( parent );

  if ( !parentMedium ) {
    if ( media.size() > 0 )
      return true;

    return false;
  }

  return parentMedium->hasChildren();
}

bool MediaModel::setData ( const QModelIndex& index, const QVariant& value, int role )
{
  if ( role != Qt::EditRole )
    return false;

  Medium *medium = mediumFromIndex ( index );

  if ( !medium )
    return false;

  return medium->setValue ( index.column(), value );
}

QVariant MediaModel::headerData ( int section, Qt::Orientation orientation, int role ) const
{
  if ( ( role != Qt::DisplayRole ) && ( role != Qt::EditRole ) )
    return QVariant();

  if ( orientation == Qt::Vertical ) {
    return section + 1;
  }

  if ( ( section < 0 ) || ( headerList.size() <= section ) )
    return QVariant();

  return headerList[section];
}

Qt::ItemFlags MediaModel::flags ( const QModelIndex& index ) const
{
	return QAbstractItemModel::flags ( index );	
}

void MediaModel::sort ( int column, Qt::SortOrder order )
{
  QAbstractItemModel::sort ( column, order );
}

bool MediaModel::clear()
{
//    emit beginResetModel();
  emit layoutAboutToBeChanged();
  qDeleteAll ( media );
  media.clear();
  emit layoutChanged();
//    emit endResetModel();

  if ( media.size() == 0 )
    return true;

  return false;
}

void MediaModel::searchOtherUsers(bool lastStepResult)
{
	thread->disconnect(this, SLOT(searchOtherUsers(bool)));
	if ( (rowCount() == 0) || (!lastStepResult) || (lastUserId == -2)) {
		ProgressStatusWidget *statusWidget = new ProgressStatusWidget();
		statusWidget->setLabel(tr("Local (All): "));
		connect(statusWidget, SIGNAL(removeStatusWidget(QWidget*)), this, SLOT(removeStatusWidgetByPass(QWidget*)));
		connect(thread, SIGNAL(setRange(int,int)), statusWidget, SLOT(setRange(int,int)));
		connect(thread, SIGNAL(setValue(int)), statusWidget, SLOT(setValue(int)));
		connect(thread, SIGNAL(finished(bool)), statusWidget, SLOT(removeMe()));
		emit showStatusWidget(statusWidget);
		// search for local media of other users and afterwards search searchplugins
		connect(thread, SIGNAL(finished(bool)), this, SLOT(searchSearchPlugins(bool)));
		
		QStringList notUserConstraints = lastLoad.strings;
		QList<QVariant> notUserConstraintBindValues = lastLoad.bindValues;
		if ( lastLoad.actualUserId != 0 ){
			if(!lastLoad.userNotNeeded){
				notUserConstraints.append ( QString ( " usersId<>?" ) );
				notUserConstraintBindValues << lastLoad.actualUserId;
			}
		}

		QString where;

		if ( notUserConstraints.size() > 0 )
			where = " WHERE " + notUserConstraints.join ( " AND " );

		thread->execQuery(QString ( "SELECT * FROM %1%2 %3" ).arg(lastLoad.table).arg ( where ).arg(lastLoad.order), notUserConstraintBindValues);

		// if we have no media found, and searchExternal is true, search external sources
		emit removeStatusWidget(statusWidget);
	}
	else{
		searchSearchPlugins(true);
	}
}

void MediaModel::searchSearchPlugins(bool lastStepResult)
{
	thread->disconnect(this, SLOT(searchSearchPlugins(bool)));
	if(!lastLoad.searchExternal)
		return;
	if ( (rowCount() == 0) || (!lastStepResult) || (lastUserId == -2)) {
		// evaluating search oder
		QSettings settings;
		settings.beginGroup ( "searchOrders" );
		QStringList order = settings.value ( "order" + myGroup ).toStringList();
		settings.endGroup();

		QList<SearchPlugin*> orderedPlugins;

		if ( order.size() < 1 ) {
			orderedPlugins = lastLoad.lastSearchPlugins;
		}
		else {
			foreach ( QString pluginName, order ) {
				foreach ( SearchPlugin *plugin, lastLoad.lastSearchPlugins ) {
					if ( pluginName == plugin->name() )
						orderedPlugins.append ( plugin );
				}
			}
		}
		// searching as long we should search
		foreach ( SearchPlugin *plugin, orderedPlugins ) {
			// check aborting conditions
			if ( (rowCount() != 0) && (lastUserId != -2))
				break;

			plugin->search ( this, lastLoad.rawConstraints );
		}
	}
}

void MediaModel::errorByPass ( QString msg )
{
  emit error ( msg );
}

void MediaModel::warningByPass ( QString msg )
{
  emit warning ( msg );
}

void MediaModel::informationByPass ( QString msg )
{
  emit information ( msg );
}

void MediaModel::showStatusWidgetByPass ( QWidget* widget )
{
  emit showStatusWidget ( widget );
}

void MediaModel::removeStatusWidgetByPass ( QWidget* widget )
{
  emit removeStatusWidget ( widget );
}

void MediaModel::setupModel(QueryThread *queryThread)
{
	setObjectName("MediaModel"+ QDateTime::currentDateTime().toString("yyyyMMddhhmmzzz"));
  MSqlDatabase db = MSqlDatabase::database();
	
	thread = queryThread;
	// we don't start a thread everytime here, because right now,
	// we use the thread only at the load function,
	// therefore it is initialized there
// 	if(!thread){
// 		thread = new QueryThread();
// 		ownThread = true;
// 		connect(thread, SIGNAL(started()), this, SLOT(setupThread()));
// 	}
// 	if(!thread->isStarted())
// 		thread->start();
// 	connect(thread, SIGNAL(result(QSqlRecord)), this, SLOT(addMedium(QSqlRecord)));
	
  Medium *medium = newMedium();

  // set names
  myName = medium->name() + "Model";
  myGroup = medium->group() + "Model";

  if ( medium->type() != 0 ){
    mediaCondition.setConstraint(" materialtypesId=?" );
		mediaCondition.addBindValue( medium->type() );
	}

	// set columns to hide
	headerList = medium->supportedAttributesTrans();
	QStringList hideKeys;
	hideKeys << "description" << "min" << "listPrice" << "listPriceCurrency";
	foreach(QString key, hideKeys){
		ColKeyTrans item = medium->attrs(key);
		if(item.column > -1)
			hideColumns << item.column;
	}
	biggestColumn = medium->attrs ( "title" ).column;
	delete medium;
}

void MediaModel::connectMedium ( Medium* object )
{
  connect ( object, SIGNAL ( error ( QString ) ), this, SLOT ( errorByPass ( QString ) ) );
  connect ( object, SIGNAL ( warning ( QString ) ), this, SLOT ( warningByPass ( QString ) ) );
  connect ( object, SIGNAL ( information ( QString ) ), this, SLOT ( informationByPass ( QString ) ) );

  connect ( object, SIGNAL ( showStatusWidget ( QWidget* ) ), this, SLOT ( showStatusWidgetByPass ( QWidget* ) ) );
  connect ( object, SIGNAL ( removeStatusWidget ( QWidget* ) ), this, SLOT ( removeStatusWidgetByPass ( QWidget* ) ) );
}

void MediaModel::itemChanged(Medium* med)
{
  emit dataChanged ( indexFromMedium(med), index ( indexFromMedium(med).row(), columnCount() ) );
}

void MediaModel::setupThread()
{
	MSqlDatabase db = MSqlDatabase::database();
	thread->dbSetup(db.databaseName(), db.userName(), db.password(), db.hostName(), db.driverName());
}

void MediaModel::doAddMedium(Medium* med)
{
	if(!med)
		return;

	med->setCompatibilityMode(compatibilityMode);

  connect ( med, SIGNAL ( changed(Medium*) ), this, SLOT ( itemChanged(Medium*) ) );
  emit layoutAboutToBeChanged();
  connectMedium ( med );
  media.append ( med );
  emit layoutChanged();
}

#include "mediamodel.moc"
