/*
    Meganizer - Media Organzier is a media management software
    Copyright (C) 2010  EngSaS - Engineering Solutions and Services Langenbach. All rights reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MEDIUM_H
#define MEDIUM_H

#include "libmacro.h"

class MDialog;
class MWidget;
class MaterialPersonsModel;
class MINModel;
class EditMedium;

#include <mobject.h>
#include <QString>
#include <QList>
#include <QVariant>
#include <QStringList>
#include <QPixmap>
#include <QSqlRecord>
#include <QModelIndex>
#include <QIcon>

class QNetworkReply;
class QTextTable;
class QStringListModel;

class ColKeyTrans {
	public:
		ColKeyTrans(int col = -1, QString key = QString(), QString trans = QString()){
			column = col;
			this->key = key;
			translation = trans;
		}
		int column;
		QString key;
		QString translation;
};
		
/**
	* @brief This is the base of all items in models which handle media.
	*
	* To add new media types, subclass this class and also the MediaModel class.
	* To do so, please make sure, that you respect the compatibilityMode, so that
	* your subclass can be shown quite well in the pure MediaModel. See compatibilityMode()
	* for details.
	*/
class EXPORT_DLL Medium : public MObject
{
  Q_OBJECT
  // don't know how to localize that
//   Q_ENUMS(ExportFormat)
  public:
		enum Origin { Local, Manual, SearchPlugin };
		Q_DECLARE_FLAGS(Origins, Origin);
		
		/**
			* Table means, all media are put in one table
			* TablePerType means, there is one table per media type
			* Long means, one page per media
			*/
		enum ExportFormat { Table };
// 		enum ExportFormat { Table, TablePerType, Long };
		static QString exportFormatToString(ExportFormat format);
		
    Medium ( Medium *newParent = 0 );
		Medium ( QString name, QString group, Medium *newParent = 0 );
    virtual ~Medium();
		
		virtual Medium *newMedium();
		QString name();
    QString group();
    /**
    	* this function fetches neeeded stuff from database
    	* @param searchId id of table materials to indicate which item should be fetched
    	* @attention A subclass of Medium should implement it's own select function
    	* @attention to select the additional stuff which the subclass needs
    	* @attention the base class implementation can be called out of the subclassed select
    	* @attention function to fetch the standard stuff
    	*/
    virtual bool select ( unsigned int searchId, unsigned int userId = 0 );
    /**
      * @todo transfer the loading of materialpersons into the materialpersonsmodel
      */
    virtual bool select( QSqlRecord record);
    /**
      * the function stores the data of a medium into the database
      * @param addOwner if called as true, an entry at userownsmaterial with actual user
      * @param addOwner is added, if the actual user doesn't own it already
      * @attention addOwner takes only effect if an update is done, 
      * @attention otherwise the owner is set to actual user
      */
    virtual bool store(bool addOwner = false, bool startTransaction = true);
		
    /**
    	* @todo check wether QVariant can be converted to needed value type
			* @attention if submit persons like authors (,...),
			* @attention you need this triple as QStringList: Name, givenName (may be empty), roleName or Id (may be empty)
			* @attention if you want to set MINs, you can either give only a min and the type will be gassed,
			* @attention or you can give a QStringList: min, type
			* in the subclass function, only special attributes, which Medium does not support, must be handled
			* to handle standard stuff, call return Medium::setValue(QString, QVariant) at the end
			* Please make sure, that you support the compatibilityMode() like this
			* @code
			*   if(compatibilityMode())
			*     return Medium::supportedAttrs();
			*   // Do what you need, if compatibilityMode() is false
			* @endcode
    	*/
		virtual bool setValue ( QString attribute, QVariant value );
    bool setValue ( int attributeIndex, QVariant value );
		/**
			* Please make sure, that you support the compatibilityMode() like this
			* @code
			*   if(compatibilityMode())
			*     return Medium::supportedAttrs();
			*   // Do what you need, if compatibilityMode() is false
			* @endcode
			*/
		virtual QVariant value ( QString attribute, int role );
    QVariant value ( int attributeIndex, int role );
    /**
    	* a list with supported attributes (not translated)
    	*/
    QStringList supportedAttributes();
		QStringList supportedAttributesTrans();
		/**
			* @brief A list of supported attributes (not translated, translated).
			*
			* Please make sure, that you support the compatibilityMode() like this
			* @code
			*   if(compatibilityMode())
			*     return Medium::supportedAttrs();
			*   // Do what you need, if compatibilityMode() is false
			* @endcode
			*/
		virtual QList<ColKeyTrans> supportedAttrs();
		/**
			* @brief returns the ColKeyTrans belonging to attribut @p key.
			*
			* It can be used to insert translations of this class into the
			* supportedAttrs list of a subclass by
			* attrsSubclass << Medium::attrs(key, newCol);
			* @return if key not an empty ColKeyTrans is returned
			* @note if newCol is greator -1 the base class implementation is used!
			*/
		ColKeyTrans attrs(QString key, int newCol = -1);


    virtual EditMedium* editWidget(MWidget* parent = 0);
		/**
			* Do a bit more around Medium::editWidget
			* @param editMode is EditMedium::EditMode enum
			*/
		virtual bool edit(int editMode, MWidget *parent = 0);
		/**
			* checks wether the medium has the minimum needed stuff
			*/
		virtual bool isValid();
		/**
			* The sublass should call this implementation with
			* Medium::copyMedium() and copy specific properties
			* afterwards
			*/
		virtual Medium* copyMedium();
		virtual bool exportMedium(QTextTable *table, Medium::ExportFormat format, bool titleOnly = false);

    void setParent ( Medium *newParent );
    Medium *parent() const;
    void setId ( unsigned int newId );
    unsigned int id() const;
    void setType ( unsigned int typeId );
    unsigned int type() const;
		QString typeName() const;
		/**
			* sets the id of the manufacturer at the database to manuId
			* and tries to get the name of the manufacturer and sets it to the result
			* @param manuId the id of the manufacturer at the manufacturers table
			* @param setName wether the corresponding name to the id should be setted or not (if possible)
			*/
    void setManufacturer ( unsigned int manuId, bool setName = true );
		/**
			* the manufacturer name is set to name
			* if setId is true, it tries to find the id corresponding to the name at the db and sets it
			*/
		void setManufacturer ( QString name, bool setId = true );
		/**
			* returns the id of the manufacturer
			*/
    unsigned int manufacturer() const;
		/**
			* returns the name of the manufacturer
			*/
		QString manufacturerName() const;
		bool prepareManufacturer();
    void setGenre ( unsigned int genreId, bool setName = true );
		void setGenre ( QString name, bool setId = true );
    unsigned int genre() const;
		QString genreName() const;
		bool prepareGenres();
    void setTitle ( QString newTitle );
    QString title() const;
    void setMIN ( QString newMin, QString minType = QString() );
		void setMIN ( QString newMin, unsigned int minTypeId );
		void setMIN ( MINModel *newMINModel );
    MINModel *mins() const;
		QString minsCombined() const;
		/**
			* @return the min, which's type is into @p minTypes
			* @return if more than one MIN's of @p minTypes are found, 
			* @return the one which lowest type index (index of type at @p minTypes) is returned
			* @return if no MIN with wished type is found or no other type can be converted to, empty QString is returned
			*/
		QString min(QStringList minTypes) const;
		QString min(QString minType) const;
    void setListPrice ( double price );
    double listPrice() const;
    void setListPriceCurrency ( unsigned int curId );
		/**
			* @p shortCode can be the shortcode of a currency or the name
			*/
		void setListPriceCurrency ( QString shortCode );
    unsigned int listPriceCurrency() const;
    void setDescription(QString desc);
    QString description() const;
    void setImage ( QPixmap newImage );
		void setImage ( QString url );
    QPixmap image() const;
		QByteArray imageToByteArray();
    void setUser ( unsigned int userId );
    unsigned int user() const;
    void setBoughtPrice ( double price );
    double boughtPrice() const;
    void setBoughtPriceCurrency ( unsigned int curId );
		/**
			* @see setListPriceCurrency()
			*/
		void setBoughtPriceCurrency ( QString shortCode );
    unsigned int boughtPriceCurrency() const;
    void setEvaluation ( int eval );
    int evaluation() const;
    void setWish ( bool isWish );
    bool wish() const;
		void setLiteratureOnly ( bool isLiterature );
		bool literatureOnly () const;
		void setGiftOnly( bool isGift );
		bool giftOnly() const;
		void setSourceIcon(QIcon icon);
		QIcon sourceIcon() const;
		Medium::Origin source() const;
		void setSource(Medium::Origin newOrigin);
    
    // handle material persons
		void setMaterialPerson(QString name, QString givenName, QString role);
		void setMaterialPerson(QString name, QString givenName, unsigned int role);
		void setMaterialPersons(MaterialPersonsModel *personModel);
    MaterialPersonsModel *materialPersons() const;
		QString materialPersonsCombined() const;
    bool removeMaterialPerson(QModelIndex index);
		void setMarkForRemovel();
    bool markForRemovel();

		// handle keywords
		void addKeyword(QString keyword);
		bool removeKeyword(QString keyword);
		/**
			* @returns the keywords as a QStringListModel
			* You must take care to delete the model of your own
			*/
		QStringListModel *keywords();
		
    void setChild ( Medium *newChild );
    QList<Medium*> childs() const;
    int rowCount() const;
    int columnCount();
    bool hasChildren() const;
		void setCompatibilityMode(bool enable);
		
    bool operator== (const Medium &med) const;
		bool operator!= (const Medium &med) const;

  protected:
		/**
			* Use this function to check wether compatibility mode
			* is enabled or not. Compatibility mode is used, to show
			* a subclass of this class in a pure MediaModel. Therefore
			* the column order and hiding is different to the subclasses.
			* You should respect this espacially at this functions:
			*  - value(QString, int), setValue(QString, QVariant)
			*  - supportedAttrs().
			*
			* Usually you should only call the base implementation in your
			* functions, if compatibilityMode() is set to true.
			*/
		bool compatibilityMode() const;
		
		bool transactionsSupported;
		QString myGroup, myName;

	private slots:
		/**
			* @todo seems to procude a segfault from time to time
			*/
		void imageDownloaded();
		
  private:
		void init(Medium *newParent = 0);
		bool loadKeywords();
		bool storeKeywords();
		
    Medium *myParent;
    unsigned int myId, myTypeId, myManuId, myGenreId, myCurId, myUserId, myBoughtCurId;
    QString myTypeName, myTitle, myEan, myDescription, myManuName, myGenreName;
    double myListPrice, myBoughtPrice;
    int myEval;
    bool myWish, myLiterature, myGift, shouldBeRemoved;
		bool myCompatibilityMode;
    QPixmap myImage;
    QList<Medium*> myChilds;
    MaterialPersonsModel *matPersonModel;
		MINModel *minModel;
		QNetworkReply *reply;
		QIcon mySourceIcon;
		Medium::Origin myOrigin;
		QStringList keywordList;
    
  signals:
    void changed(Medium *medium);
};

Q_DECLARE_OPERATORS_FOR_FLAGS(Medium::Origins)

#endif // MEDIUM_H
