/*
    Meganizer - Media Organzier is a media management software
    Copyright (C) 2010  EngSaS - Engineering Solutions and Services Langenbach. All rights reserved.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef SQLINIT_H
#define SQLINIT_H

#include "libmacro.h"

#include "sqltable.h"
#include "sqlfield.h"
#include "mobject.h"
#include "msqldatabase.h"

/**
  * provides basics, so that a db structure can easily be checked
  * and created or modified if needed
  * @todo error message when inserting of vaules have failed
  * @todo check needed rights and ask for username and password with extended rights, if they are needed
  *       for example things like create tables, ...
  */
class EXPORT_DLL SqlInit : public MObject
{
  Q_OBJECT
  public:
    /**
    	* possible SQL permissions
    	* CREATE assumes ALTER and CREATE VIEW
    	*/
    enum privileges { SELECT, INSERT, UPDATE, DELETE, CREATE, ALTER };
		enum Status { BadSettings, LoggedIn, LoginFailed, UpdateNeeded, UpdateFailed, NoUpdateNeeded };

    SqlInit ( QObject *parent = 0,
							QString username = QString(), QString host = QString(),
							QString database = QString(), QString driver = QString(),
							QString password = QString());
    ~SqlInit();
    /**
      * logs in and after login it's check the structure
      * of the database against the given one at the scheme() function
    	* @param runUpdate is set to false login returns true if an update is needed
    	* @param runUpdate is set to true, login runs the update and return true if the updates are done successfull
      */
    SqlInit::Status login ( bool loginOnly = true, bool runUpdate = false );

  protected:
    /**
      * must be implement by all subclasses
      * it stores the db structure which is needed
      */
    virtual void scheme() = 0;
		/**
			* in this routine, the subclass can do additional stuff on upgrading from one  version to another
			* e.g. moving data around to fit the old data to new layout and so on
			* @param fromVersion the version which should be updated
			* @param toVersion the version which should be reached (0 means update to newest, default)
			*/
		bool runUpdateRoutines(int fromVersion, QString& errorString, int toVersion);
		/**
			* used to add the updateroutine to the map
			* @param toVersion Version which is reached, after the updateroutine is run
			* @note the updateroutine should handle a QString& to give back error messages @see DBInit::updateFrom3To4()
			* @note and it get's the own SqlInit name at first parameter so this is a template:
			* @note bool myUpdateRoutine(QString thisName, QString &errorString);
			*/
		void addUpdateRoutine(int toVersion, bool (*functionPointer)(QString, QList<SqlTable*>, QString&) );

		/**
			* adds a new table to the list
			* @param tableName the name of the table
			* @param appendFieldId if true (default) the standard id field is added
			* @return pointer to the table object
			*/
    SqlTable *addTable ( QString tableName, bool appendFieldId = true );
    /**
      * initializes the structure by calling scheme()
      */
    void init();

    /**
      * current version of db structure
      */
    int currentVersion;
    /**
      * prefix of all tables which belongs to this module
      */
    QString thisName;

    QString moduleLabel;
    /**
      * a list of all tables of the database which are needed by this module
      */
    QList<SqlTable*> tables;
    QList<QStringList> views;
		MSqlDatabase db;
		/**
			* the first int (QMap-Key) is the result version, which is reached, if updateFunction runs successfully. 
			* updateFunction should than return true
			* @see DBInit for examples
			*/
		QMap<int, bool (*)(QString, QList<SqlTable*>, QString&)> updateRoutines;
// 		QMap<int, SqlInit::updateRoutine> updateRoutines;

  private slots:
    /**
      * initializes the update
      */
    bool updateDb();

  private:
    /**
      * checks wether update of the structure is needed or not
      * it also checks wether structure is newer than needed
      * which might be bad too
      */
    bool updateNeeded();
    /**
      * logs into the database
      */
    void doLogin();
    /**
      * does the update
      */
    bool doUpdate();
    /**
      * compares to tables
      * in mode @p create = false it returns true if both
      * tables are identical and false if not
      * if @p update is set to true, it updates the table or even create it.
			* @note Fields are not deleted here, to give you the chance to do something with the
			*   data, which is in the deleted fields. Therefore, you must delete fields in an
			*   updateRoutine. Also default values are not touched, if fields are deleted to avoid
			*   errors (because fields are not deleted, but default values needs the new structure).
      */
    bool compareTables ( SqlTable *table, QString &errorString, bool create = false );

    /**
      * checks, if view with current statement exists
      * @param view the view which a create statement should be created
      * @param error a pointer to a QString where the error message is appended
      * @param create create view if it not exist or with other select statement
      * @return true if view already exists and is the same as needed
			* @todo Because, checking is quite complicated, the views are dropped and created again, this should be improved!
      */
    bool compareViews ( QStringList view, QString &errorString, bool create = false );

    /**
      * creates the SQL statements which are needed to create
      * the table
      * @param table the table which should be created
      * @return SQL query string which can be used to create the table
      */
    QString createTableStatement ( SqlTable *table );
    
    /**
      * creates the SQL statement to create the field
      * e.g. a Field with name cell2 and varchar(20) not null
      * will result in this statement:
      * cell2 varchar(20) not null
      */
    QString fieldStatement(SqlField *field);

    /**
      * create views statement
      * @param view index of view at QList views;
      * @return SQL query to create view
      */
    QString createViewStatement ( QStringList view );

    QString dbName, versionTable;
    bool transactions;
    bool sqLite;
    QList<SqlInit::privileges> neededPrivs;
		int actVersion;
		QString username, password, host, driver;

  signals:
    void actualStep ( QString msg );
    void finishedStep();
    void neededSteps ( int num );
};

#endif // SQLINIT_H
